#include <stdlib.h>
#include "BinHeap.h"


static int binHeapCmp(BinHeapT *heap, const void *data1, const void *data2)
{
    if (heap->heapType == BIN_HEAP_TYPE_MIN)
    {
        return heap->compareFunction(data1, data2);
    }
    else
    {
        return -heap->compareFunction(data1, data2);
    }
}

BinHeapT *binHeapCreate(BinHeapType heapType,
                        BinHeapCompareFunc compareFunction)
{
    BinHeapT *heap;

    heap = malloc(sizeof(BinHeapT));

    if (heap == NULL)
    {
        return NULL;
    }

    heap->heapType = heapType;
    heap->count = 0;
    heap->compareFunction = compareFunction;

    /* Initial size of 16 elements */

    heap->size = 16;
    heap->values = malloc(sizeof(void *) * heap->size);

    if (heap->values == NULL)
    {
        free(heap);
        return NULL;
    }

    return heap;
}

void binHeapFree(BinHeapT *heap)
{
    free(heap->values);
    free(heap);
}

int binHeapInsert(BinHeapT *heap, void *value)
{
    void **newValues;
    int index;
    int newsize;
    int parent;

    /* Possibly realloc the heap to a larger size */
    if (heap->count >= heap->size)
    {   /* Double the table size */
        newsize = heap->size * 2;
        newValues = realloc(heap->values, sizeof(void *) * newsize);
        if (newValues == NULL)
        {
            return 0;
        }
        heap->size = newsize;
        heap->values = newValues;
    }
    /* Add to the bottom of the heap and start from there */
    index = heap->count;
    ++heap->count;

    /* Percolate the value up to the top of the heap */
    while (index > 0)
    {   /* The parent index is found by halving the node index */
        parent = (index - 1) / 2;
        /* Compare the node with its parent */
        if (binHeapCmp(heap, heap->values[parent], value) < 0)
        {   /* Ordered correctly - insertion is complete */
            break;
        }
        else
        {   /* Need to swap this node with its parent */
            heap->values[index] = heap->values[parent];
            /* Advance up to the parent */
            index = parent;
        }
    }
    /* Save the new value in the final location */
    heap->values[index] = value;
    return 1;
}

void *binHeapPop(BinHeapT *heap)
{
    void *result;
    void *newValue;
    int index;
    int nextIndex;
    int child1, child2;

    /* Empty heap? */

    if (heap->count == 0)
    {
        return BIN_HEAP_NULL;
    }
    /* Take the value from the top of the heap */
    result = heap->values[0];
    /* Remove the last value from the heap; we will percolate this down
     * from the top. */
    newValue = heap->values[heap->count - 1];
    --heap->count;
    /* Percolate the new top value down */
    index = 0;
    for (;;)
    {
        /* Calculate the array indexes of the children of this node */
        child1 = index * 2 + 1;
        child2 = index * 2 + 2;
        if (child1 < heap->count
               && binHeapCmp(heap, newValue, heap->values[child1]) > 0)
        {   /* Left child is less than the node.  We need to swap
             * with one of the children, whichever is less. */
            if (child2 < heap->count
                && binHeapCmp(heap, heap->values[child1], heap->values[child2]) > 0)
            {
                nextIndex = child2;
            }
            else
            {
                nextIndex = child1;
            }
        }
        else if (child2 < heap->count
                 && binHeapCmp(heap, newValue, heap->values[child2]) > 0)
        {   /* Right child is less than the node.  Swap with the
             * right child. */
            nextIndex = child2;
        }
        else
        {   /* Node is less than both its children.  The heap condition
             * is satisfied.  We can stop percolating down. */
            heap->values[index] = newValue;
            break;
        }
        /* Swap the current node with the least of the child nodes. */
        heap->values[index] = heap->values[nextIndex];
        /* Advance to the child we chose */
        index = nextIndex;
    }
    return result;
}

int binHeapGetCount(BinHeapT *heap)
{
    return heap->count;
}

