/**
 * @file BinHeap.h
 *
 * @brief Binary heap.
 *
 * A binary heap is a heap data structure implemented using a
 * binary tree.  In a heap, values are ordered by priority.
 *
 * To create a binary heap, use @ref binHeapCreate.  To destroy a
 * binary heap, use @ref binHeapFree.
 *
 * To insert a value into a binary heap, use @ref binHeapInsert.
 *
 * To remove the first value from a binary heap, use @ref binHeapPop.
 *
 */

#ifndef BIN_HEAP_H
#define BIN_HEAP_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Heap type.  If a heap is a min heap (@ref BIN_HEAP_TYPE_MIN), the
 * values with the lowest priority are stored at the top of the heap and
 * will be the first returned.  If a heap is a max heap
 * (@ref BIN_HEAP_TYPE_MAX), the values with the greatest priority are
 * stored at the top of the heap.
 */

typedef enum {
    /** A minimum heap. */
    BIN_HEAP_TYPE_MIN,
    /** A maximum heap. */
    BIN_HEAP_TYPE_MAX
}
BinHeapType;

/**
 * A null @ref BinaryHeapValue.
 */

#define BIN_HEAP_NULL ((void *) 0)

/**
 * Type of function used to compare values in a binary heap.
 *
 * @param value1           pointer to the first value.
 * @param value2           pointer to the second value.
 * @return                 A negative number if value1 is less than value2,
 *                         a positive number if value1 is greater than value2,
 *                         zero if the two are equal.
 */

typedef int (*BinHeapCompareFunc)(const void *value1,
                                  const void *value2);

/**
 * A binary heap data structure.
 */

struct _BinaryHeap
{
    BinHeapType heapType;
    void **values;
    int count;
    int size;
    BinHeapCompareFunc compareFunction;
};

typedef struct _BinaryHeap BinHeapT;

/**
 * Create a new @ref BinaryHeap.
 *
 * @param heapType        The type of heap: min heap or max heap.
 * @param compareFunction  Pointer to a function used to compare the priority
 *                         of values in the heap.
 * @return                 A new binary heap, or NULL if it was not possible
 *                         to allocate the memory.
 */
BinHeapT *binHeapCreate(BinHeapType heapType,
                        BinHeapCompareFunc compareFunction);

/**
 * Destroy a binary heap.
 *
 * @param heap             The heap to destroy.
 */
void binHeapFree(BinHeapT *heap);

/**
 * Insert a value into a binary heap.
 *
 * @param heap             The heap to insert into.
 * @param value            The value to insert.
 * @return                 Non-zero if the entry was added, or zero if it
 *                         was not possible to allocate memory for the new
 *                         entry.
 */
int binHeapInsert(BinHeapT *heap, void  *value);

/**
 * Remove the first value from a binary heap.
 *
 * @param heap             The heap.
 * @return                 The first value in the heap, or
 *                         @ref BIN_HEAP_NULL if the heap is empty.
 */
void *binHeapPop(BinHeapT *heap);

/**
 * Find the number of values stored in a binary heap.
 *
 * @param heap             The heap.
 * @return                 The number of values in the heap.
 */
int binHeapGetCount(BinHeapT *heap);

#ifdef __cplusplus
}
#endif

#endif /* #ifndef BIN_HEAP_H */


